# frozen_string_literal: true

require 'drillbit/resource/naming'
require 'drillbit/resource/model'

module  Drillbit
module  AuthorizableResource
  RESOURCE_COLLECTION_ACTIONS = %w{index}.freeze

  module ClassMethods
    def authorizer_prefix
      @authorizer_prefix ||= name[Resource::Naming::CONTROLLER_RESOURCE_NAME_PATTERN, 2]
    end

    def authorizer_class
      @authorizer_class ||= "#{authorizer_prefix}" \
                            "Authorizers::" \
                            "#{resource_class_name}"
                              .constantize
    rescue NameError
      'Drillbit::Authorizers::Query'.constantize
    end

    def authorizer_scope_class
      @authorizer_scope_class ||= "#{authorizer_prefix}" \
                                  "Authorizers::" \
                                  "#{resource_class_name}" \
                                  "::Scope"
                                    .constantize
    rescue NameError
      'Drillbit::Authorizers::Scope'.constantize
    end

    def authorizer_resource_params_class
      @authorizer_resource_params_class ||= "#{authorizer_prefix}" \
                                            "Authorizers::" \
                                            "#{resource_class_name}" \
                                            "::ResourceParameters"
                                              .constantize
    rescue NameError
      'Drillbit::Authorizers::Parameters::Resource'.constantize
    end

    def authorizer_filtering_params_class
      @authorizer_filtering_params_class ||= "#{authorizer_prefix}" \
                                             "Authorizers::" \
                                             "#{resource_class_name}::" \
                                             "FilteringParameters"
                                               .constantize
    rescue NameError
      'Drillbit::Authorizers::Parameters::Filtering'.constantize
    end

    def authorizer_inclusions_params_class
      @authorizer_inclusions_params_class ||= "#{authorizer_prefix}" \
                                              "Authorizers::" \
                                              "#{resource_class_name}::" \
                                              "InclusionParameters"
                                                .constantize
    rescue NameError
      'Drillbit::Authorizers::Parameters::Inclusions'.constantize
    end
  end

  def self.included(base)
    base.include Resource::Naming
    base.extend  ClassMethods

    base.before_action :authorize
  end

  private

  def authorize
    return if authorizer.public_send(authorization_query)

    Erratum.raise(
      'ForbiddenError',
      resource_name: self.class.singular_resource_name,
      resource_id:   [params[:id]],
      action:        action_name,
    )
  end

  def authorizer
    @authorizer ||= self
                      .class
                      .authorizer_class
                      .new(action:   action_name,
                           token:    token,
                           user:     authorized_user,
                           issuer:   authorized_issuer,
                           params:   authorized_params,
                           resource: authorized_resource)
  end

  def authorized_scope
    @authorized_scope ||= self
                            .class
                            .authorizer_scope_class
                            .new(action:     action_name,
                                 token:      token,
                                 user:       authorized_user,
                                 issuer:     authorized_issuer,
                                 params:     authorized_params,
                                 scope_root: authorized_scope_root)
                            .call
  end

  def authorized_inclusions
    @authorized_inclusions ||= self
                                 .class
                                 .authorizer_inclusions_params_class
                                 .new(action: action_name,
                                      token:  token,
                                      user:   authorized_user,
                                      issuer: authorized_issuer,
                                      params: authorized_params)
                                 .call
  end

  def authorized_params
    @authorized_params ||= authorizer_params_class
                             .new(action: action_name,
                                  token:  token,
                                  user:   authorized_user,
                                  issuer: authorized_issuer,
                                  params: params)
                             .call
  end

  # rubocop:disable Metrics/AbcSize, Metrics/PerceivedComplexity
  # rubocop:disable Metrics/CyclomaticComplexity, Metrics/MethodLength
  # rubocop:disable Metrics/BlockNesting
  def authorized_attributes
    @authorized_attributes ||= begin
      attributes             = authorized_params
                                 .fetch(:data,       {})
                                 .fetch(:attributes, authorized_params.class.new)

      relationships = authorized_params.class.new

      authorized_params
        .fetch(:data,          {})
        .fetch(:relationships, authorized_params.class.new)
        .each_pair do |name, relationship|
        if relationship[:data].is_a?(Array)
          if (relationship[:data][0] || {})[:attributes]
            relationships["#{name}_attributes"] = relationship[:data].map do |datum|
              attrs                             = datum[:attributes].dup

              attrs.delete(:__id__)
              attrs[:id] = datum[:id] if datum[:id]

              attrs
            end
          else
            attribute = "#{Drillbit::Utilities::String.singularize(name)}_ids".to_sym

            relationships[attribute] = relationship[:data].map { |datum| datum[:id] }
          end
        elsif relationship[:data].nil? || relationship[:data].is_a?(Hash)
          attribute = name.to_sym

          relationships[attribute] = relationship[:data][:id]
        end
      end

      relationships.permit! if relationships.respond_to?(:permit!)

      ActiveSupport::Deprecation.silence do
        attributes.merge(relationships)
      end
    end
  end
  # rubocop:enable Metrics/BlockNesting
  # rubocop:enable Metrics/CyclomaticComplexity, Metrics/MethodLength
  # rubocop:enable Metrics/AbcSize, Metrics/PerceivedComplexity

  def authorized_resource
    return if RESOURCE_COLLECTION_ACTIONS.include?(action_name)

    @authorized_resource ||= public_send(self.class.singular_resource_name)
  end

  def authorized_collection
    return unless RESOURCE_COLLECTION_ACTIONS.include?(action_name)

    @authorized_collection ||= \
      Resource::Model
        .new(resource:   public_send(self.class.plural_resource_name),
             parameters: authorized_params)
  end

  def authorizer_params_class
    @authorizer_params_class ||= if RESOURCE_COLLECTION_ACTIONS.include?(action_name)
                                   self.class.authorizer_filtering_params_class
                                 else
                                   self.class.authorizer_resource_params_class
                                 end
  end

  def authorized_scope_root
    @authorized_scope_root ||= "#{self.class.authorizer_prefix}" \
                               "#{self.class.resource_class_name}"
                                 .constantize
  end

  def authorized_user
    current_user
  end

  def authorized_issuer
    current_issuer
  end

  def authorization_query
    @authorization_query ||= "able_to_#{action_name}?"
  end
end
end
